"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MultiStatusResponseBuilder = exports.MultiStatusResponse = exports.ActionDestinationErrorResponse = exports.ActionDestinationSuccessResponse = exports.Action = exports.hookTypeStrings = void 0;
const events_1 = require("events");
const create_request_client_1 = __importDefault(require("../create-request-client"));
const mapping_kit_1 = require("../mapping-kit");
const fields_to_jsonschema_1 = require("./fields-to-jsonschema");
const fetch_1 = require("../fetch");
const types_1 = require("./types");
const request_client_1 = require("../request-client");
const schema_validation_1 = require("../schema-validation");
const errors_1 = require("../errors");
const remove_empty_values_1 = require("../remove-empty-values");
const get_1 = require("../get");
exports.hookTypeStrings = ['onMappingSave', 'retlOnMappingSave'];
const isSyncMode = (value) => {
    return types_1.syncModeTypes.find((validValue) => value === validValue) !== undefined;
};
const INTERNAL_HIDDEN_FIELDS = ['__segment_internal_sync_mode', '__segment_internal_matching_key'];
const removeInternalHiddenFields = (mapping) => {
    return Object.keys(mapping).reduce((acc, key) => {
        return INTERNAL_HIDDEN_FIELDS.includes(key) ? acc : { ...acc, [key]: mapping[key] };
    }, {});
};
class Action extends events_1.EventEmitter {
    constructor(destinationName, definition, extendRequest) {
        super();
        this.definition = definition;
        this.destinationName = destinationName;
        this.extendRequest = extendRequest;
        this.hasBatchSupport = typeof definition.performBatch === 'function';
        this.hasHookSupport = definition.hooks !== undefined;
        if (Object.keys(definition.fields ?? {}).length) {
            this.schema = fields_to_jsonschema_1.fieldsToJsonSchema(definition.fields);
        }
        if (definition.hooks) {
            for (const hookName in definition.hooks) {
                const hook = definition.hooks[hookName];
                if (hook?.inputFields) {
                    if (!this.hookSchemas) {
                        this.hookSchemas = {};
                    }
                    const castedInputFields = {};
                    for (const key in hook.inputFields) {
                        const field = hook.inputFields[key];
                        if (field.dynamic) {
                            castedInputFields[key] = {
                                ...field,
                                dynamic: true
                            };
                        }
                        else {
                            castedInputFields[key] = {
                                ...field,
                                dynamic: false
                            };
                        }
                    }
                    this.hookSchemas[hookName] = fields_to_jsonschema_1.fieldsToJsonSchema(castedInputFields);
                }
            }
        }
    }
    async execute(bundle) {
        const results = [];
        const mapping = removeInternalHiddenFields(bundle.mapping);
        let payload = mapping_kit_1.transform(mapping, bundle.data);
        results.push({ output: 'Mappings resolved' });
        payload = remove_empty_values_1.removeEmptyValues(payload, this.schema, true);
        if (this.schema) {
            const schemaKey = `${this.destinationName}:${this.definition.title}`;
            schema_validation_1.validateSchema(payload, this.schema, { schemaKey, statsContext: bundle.statsContext });
            results.push({ output: 'Payload validated' });
        }
        let hookOutputs = {};
        if (this.definition.hooks) {
            for (const hookType in this.definition.hooks) {
                const hookOutputValues = bundle.mapping?.[hookType];
                if (hookOutputValues) {
                    hookOutputs = { ...hookOutputs, [hookType]: hookOutputValues };
                }
            }
        }
        const syncMode = this.definition.syncMode ? bundle.mapping?.['__segment_internal_sync_mode'] : undefined;
        const matchingKey = bundle.mapping?.['__segment_internal_matching_key'];
        const dataBundle = {
            rawData: bundle.data,
            rawMapping: bundle.mapping,
            settings: bundle.settings,
            payload,
            auth: bundle.auth,
            features: bundle.features,
            statsContext: bundle.statsContext,
            logger: bundle.logger,
            engageDestinationCache: bundle.engageDestinationCache,
            transactionContext: bundle.transactionContext,
            stateContext: bundle.stateContext,
            audienceSettings: bundle.audienceSettings,
            hookOutputs,
            syncMode: isSyncMode(syncMode) ? syncMode : undefined,
            matchingKey: matchingKey ? String(matchingKey) : undefined
        };
        const output = await this.performRequest(this.definition.perform, dataBundle);
        results.push({ data: output, output: 'Action Executed' });
        return results;
    }
    async executeBatch(bundle) {
        if (!this.hasBatchSupport) {
            throw new errors_1.IntegrationError('This action does not support batched requests.', 'NotImplemented', 501);
        }
        const mapping = removeInternalHiddenFields(bundle.mapping);
        let payloads = mapping_kit_1.transformBatch(mapping, bundle.data);
        const batchPayloadLength = payloads.length;
        const results = [];
        if (this.schema) {
            const schema = this.schema;
            const validationOptions = {
                schemaKey: `${this.destinationName}:${this.definition.title}`,
                throwIfInvalid: false,
                statsContext: bundle.statsContext
            };
            {
                const filteredPayload = [];
                for (let i = 0; i < payloads.length; i++) {
                    const payload = remove_empty_values_1.removeEmptyValues(payloads[i], schema);
                    const isValid = schema_validation_1.validateSchema(payload, schema, validationOptions);
                    if (isValid) {
                        filteredPayload.push(payload);
                    }
                    else {
                        results[i] = {
                            response: {
                                status: 400,
                                errortype: 'INVALID_PAYLOAD',
                                errormessage: 'Invalid payload',
                                errorreporter: 'INTEGRATIONS'
                            }
                        };
                    }
                }
                payloads = filteredPayload;
            }
        }
        let hookOutputs = {};
        if (this.definition.hooks) {
            for (const hookType in this.definition.hooks) {
                const hookOutputValues = bundle.mapping?.[hookType];
                if (hookOutputValues) {
                    hookOutputs = { ...hookOutputs, [hookType]: hookOutputValues };
                }
            }
        }
        if (payloads.length === 0) {
            return [];
        }
        if (this.definition.performBatch) {
            const syncMode = this.definition.syncMode ? bundle.mapping?.['__segment_internal_sync_mode'] : undefined;
            const matchingKey = bundle.mapping?.['__segment_internal_matching_key'];
            const data = {
                rawData: bundle.data,
                rawMapping: bundle.mapping,
                settings: bundle.settings,
                audienceSettings: bundle.audienceSettings,
                payload: payloads,
                auth: bundle.auth,
                features: bundle.features,
                statsContext: bundle.statsContext,
                logger: bundle.logger,
                engageDestinationCache: bundle.engageDestinationCache,
                transactionContext: bundle.transactionContext,
                stateContext: bundle.stateContext,
                hookOutputs,
                syncMode: isSyncMode(syncMode) ? syncMode : undefined,
                matchingKey: matchingKey ? String(matchingKey) : undefined
            };
            const requestClient = this.createRequestClient(data);
            const performBatchResponse = await this.definition.performBatch(requestClient, data);
            if (performBatchResponse instanceof fetch_1.Response && performBatchResponse.status !== 207) {
                let parsedBody = {};
                try {
                    parsedBody =
                        performBatchResponse.data ??
                            performBatchResponse.content;
                }
                catch (e) {
                    parsedBody = {};
                }
                this.fillMultiStatusResponse({
                    results,
                    batchPayloadLength,
                    status: performBatchResponse.status,
                    body: parsedBody,
                    filteredPayloads: payloads
                });
                return results;
            }
            if (performBatchResponse instanceof request_client_1.HTTPError) {
                this.fillMultiStatusResponse({
                    results,
                    batchPayloadLength,
                    status: performBatchResponse.response.status,
                    body: performBatchResponse.message,
                    filteredPayloads: payloads
                });
                return results;
            }
            if (performBatchResponse instanceof MultiStatusResponse) {
                this.setMultiStatusResponse({
                    results,
                    multiStatusResponse: performBatchResponse
                });
                return results;
            }
            this.fillMultiStatusResponse({
                results,
                batchPayloadLength,
                status: 200,
                body: {},
                filteredPayloads: payloads
            });
        }
        return results;
    }
    extractFieldContextAndHandler(field) {
        const arrayRegex = /(.*)\.\[(\d+)\]\.(.*)/;
        const objectRegex = /(.*)\.(.*)/;
        let dynamicHandlerPath = field;
        let dynamicFieldContext;
        const match = arrayRegex.exec(field) || objectRegex.exec(field);
        if (match) {
            const [, parent, indexOrChild, child] = match;
            if (child) {
                dynamicFieldContext = { selectedArrayIndex: parseInt(indexOrChild, 10) };
                dynamicHandlerPath = `${parent}.${child}`;
            }
            else {
                const parentFetcher = this.definition.dynamicFields?.[parent];
                if (parentFetcher && !(indexOrChild in parentFetcher)) {
                    dynamicHandlerPath = `${parent}.__values__`;
                    dynamicFieldContext = { selectedKey: indexOrChild };
                }
            }
        }
        return { dynamicHandlerPath, dynamicFieldContext };
    }
    async executeDynamicField(field, data, dynamicFn) {
        if (dynamicFn && typeof dynamicFn === 'function') {
            return (await this.performRequest(dynamicFn, { ...data }));
        }
        const { dynamicHandlerPath, dynamicFieldContext } = this.extractFieldContextAndHandler(field);
        const fn = get_1.get(this.definition.dynamicFields, dynamicHandlerPath);
        if (typeof fn !== 'function') {
            return Promise.resolve({
                choices: [],
                nextPage: '',
                error: {
                    message: `No dynamic field named ${field} found.`,
                    code: '404'
                }
            });
        }
        return (await this.performRequest(fn, { ...data, dynamicFieldContext }));
    }
    async executeHook(hookType, data) {
        if (!this.hasHookSupport) {
            throw new errors_1.IntegrationError('This action does not support any hooks.', 'NotImplemented', 501);
        }
        const hookFn = this.definition.hooks?.[hookType]?.performHook;
        if (!hookFn) {
            throw new errors_1.IntegrationError(`Missing implementation for hook: ${hookType}.`, 'NotImplemented', 501);
        }
        if (this.hookSchemas?.[hookType]) {
            const schema = this.hookSchemas[hookType];
            schema_validation_1.validateSchema(data.hookInputs, schema);
        }
        return (await this.performRequest(hookFn, data));
    }
    async performRequest(requestFn, data) {
        const requestClient = this.createRequestClient(data);
        const response = await requestFn(requestClient, data);
        return this.parseResponse(response);
    }
    createRequestClient(data) {
        const options = this.extendRequest?.(data) ?? {};
        return create_request_client_1.default(options, {
            afterResponse: [this.afterResponse.bind(this)],
            statsContext: data.statsContext
        });
    }
    afterResponse(request, options, response) {
        const modifiedResponse = response;
        modifiedResponse.request = request;
        modifiedResponse.options = options;
        this.emit('response', modifiedResponse);
        return modifiedResponse;
    }
    parseResponse(response) {
        if (response instanceof fetch_1.Response) {
            return response.data ?? response.content;
        }
        return response;
    }
    fillMultiStatusResponse(input) {
        const { results, batchPayloadLength, status, body, filteredPayloads } = input;
        for (let i = 0; i < batchPayloadLength; i++) {
            if (results[i]) {
                continue;
            }
            let payloadReadIndex = 0;
            results.push({
                response: {
                    status: status,
                    body: body,
                    sent: filteredPayloads ? filteredPayloads[payloadReadIndex++] : {}
                }
            });
        }
    }
    setMultiStatusResponse(input) {
        const { results, multiStatusResponse } = input;
        let resultsReadIndex = 0;
        for (let i = 0; i < multiStatusResponse.length(); i++) {
            const response = multiStatusResponse.getResponseAtIndex(resultsReadIndex++);
            if (results[i]) {
                continue;
            }
            if (response.errortype ||
                response.errormessage) {
                results[i] = {
                    response: {
                        ...response,
                        errorreporter: 'INTEGRATIONS'
                    }
                };
                continue;
            }
            results[i] = {
                response: response
            };
        }
    }
}
exports.Action = Action;
class ActionDestinationSuccessResponse {
    constructor(data) {
        this.data = data;
    }
    value() {
        return this.data;
    }
}
exports.ActionDestinationSuccessResponse = ActionDestinationSuccessResponse;
class ActionDestinationErrorResponse {
    constructor(data) {
        this.data = data;
    }
    value() {
        return this.data;
    }
}
exports.ActionDestinationErrorResponse = ActionDestinationErrorResponse;
class MultiStatusResponse {
    constructor() {
        this.responses = [];
    }
    length() {
        return this.responses.length;
    }
    pushResponse(response) {
        this.responses.push(response);
    }
    pushResponseObject(response) {
        this.responses.push(response);
    }
    pushSuccessResponse(response) {
        if (response instanceof ActionDestinationSuccessResponse) {
            this.responses.push(response);
        }
        else {
            this.responses.push(new ActionDestinationSuccessResponse(response));
        }
    }
    pushErrorResponse(response) {
        if (response instanceof ActionDestinationErrorResponse) {
            this.responses.push(response);
        }
        else {
            this.responses.push(new ActionDestinationErrorResponse(response));
        }
    }
    pushResponseObjectAtIndex(index, response) {
        this.responses[index] = response;
    }
    setSuccessResponseAtIndex(index, response) {
        if (response instanceof ActionDestinationSuccessResponse) {
            this.responses[index] = response;
        }
        else {
            this.responses[index] = new ActionDestinationSuccessResponse(response);
        }
    }
    setErrorResponseAtIndex(index, response) {
        if (response instanceof ActionDestinationErrorResponse) {
            this.responses[index] = response;
        }
        else {
            this.responses[index] = new ActionDestinationErrorResponse(response);
        }
    }
    unsetResponseAtIndex(index) {
        delete this.responses[index];
    }
    getResponseAtIndex(index) {
        return this.responses[index].value();
    }
    getAllResponses() {
        return this.responses.map((response) => response.value());
    }
}
exports.MultiStatusResponse = MultiStatusResponse;
class MultiStatusResponseBuilder {
    constructor(payloads) {
        this.payloads = payloads;
        this.invalidPayloadIndices = new Set();
        this.validPayloadIndexToBatchIndexMap = new Map();
        for (let i = 0; i < this.payloads.length; i++) {
            this.validPayloadIndexToBatchIndexMap.set(i, i);
        }
        this.multiStatusResponse = new Array(this.payloads.length);
    }
    filterPayloads(input) {
        this.invalidPayloadIndices.clear();
        this.validPayloadIndexToBatchIndexMap.clear();
        const validatedPayloads = [];
        for (let batchIndex = 0, filteredPayloadIndex = 0; batchIndex < this.payloads.length; batchIndex++) {
            if (!this.payloads[batchIndex]) {
                this.multiStatusResponse[batchIndex] = new ActionDestinationErrorResponse({
                    status: input.defaultValidationCode ?? 400,
                    errortype: 'INVALID_PAYLOAD',
                    errormessage: input.defaultValidationMessage ?? 'Invalid Payload'
                });
                continue;
            }
            const validationResponse = input.validator(this.payloads[batchIndex], batchIndex);
            if (validationResponse instanceof ActionDestinationSuccessResponse ||
                validationResponse instanceof ActionDestinationErrorResponse) {
                this.multiStatusResponse[batchIndex] = validationResponse;
                continue;
            }
            if (typeof validationResponse === 'boolean' && !validationResponse) {
                this.multiStatusResponse[batchIndex] = new ActionDestinationErrorResponse({
                    status: input.defaultValidationCode ?? 400,
                    errortype: 'INVALID_PAYLOAD',
                    errormessage: input.defaultValidationMessage ?? 'Invalid Payload'
                });
                continue;
            }
            if (input.transformer) {
                const transformedPayload = input.transformer(this.payloads[batchIndex], filteredPayloadIndex++, batchIndex);
                validatedPayloads.push(transformedPayload);
            }
            else {
                validatedPayloads.push(this.payloads[batchIndex]);
            }
            this.validPayloadIndexToBatchIndexMap.set(validatedPayloads.length - 1, batchIndex);
        }
        return validatedPayloads;
    }
    async mergeAPIResponse(input) {
        const response = await input.apiResponse;
        const parsedResponses = input.responseParser(response, this.validPayloadIndexToBatchIndexMap.size);
        if (parsedResponses.length !== this.payloads.length - this.invalidPayloadIndices.size) {
            throw new errors_1.IntegrationError(`The number of parsed responses (${parsedResponses.length}) does not match the number of valid payloads (${this.validPayloadIndexToBatchIndexMap.size}).`, 'InvalidMultiStatusResponse', 400);
        }
        if (this.invalidPayloadIndices.size === 0) {
            for (let i = 0; i < parsedResponses.length; i++) {
                this.multiStatusResponse[i] = parsedResponses[i];
            }
            return;
        }
        for (let i = 0; i < parsedResponses.length; i++) {
            const resolvedIndex = this.validPayloadIndexToBatchIndexMap.get(i);
            if (!resolvedIndex) {
                throw new errors_1.IntegrationError(`Failed to resolve MultiStatus response at index ${i}.`, 'InvalidMultiStatusResponse', 400);
            }
            this.multiStatusResponse[resolvedIndex] = parsedResponses[i];
        }
    }
    handleHTTPError(error) {
        if (this.invalidPayloadIndices.size === 0) {
            for (let i = 0; i < this.payloads.length; i++) {
                this.multiStatusResponse[i] = new ActionDestinationErrorResponse({
                    status: error.response.status,
                    errortype: 'HTTP_ERROR',
                    errormessage: error.message
                });
            }
            return;
        }
        for (const mapItem of this.validPayloadIndexToBatchIndexMap) {
            const resolvedIndex = mapItem[1];
            this.multiStatusResponse[resolvedIndex] = new ActionDestinationErrorResponse({
                status: error.response.status,
                errortype: 'HTTP_ERROR',
                errormessage: error.message
            });
        }
    }
    buildMultiStatusResponse() {
        return this.multiStatusResponse;
    }
}
exports.MultiStatusResponseBuilder = MultiStatusResponseBuilder;
//# sourceMappingURL=action.js.map